-- nvsettings.lua
-- settings for the player

local PLAYER_CFGFILE = "NuVotePlayer.cfg"

local function interpret_bool(value)
    value = value:lower()
    local translations = {
        on=true,
        off=false,
        yes=true,
        no=false,
    }
    translations['1'] = true
    translations['0'] = false
    translations["true"] = true
    translations["false"] = false
    local valid_values = {}
    for k,_ in pairs(translations) do
        table.insert(valid_values,k)
    end

    if not value:inTable(valid_values) then
        local msg = string.format("Invalid value %s. Needs to be one of [%s]",value, table.concat(valid_values,", "))
        return {
            receipt = nil,
            reason = msg
        }
    end

    return {
        receipt = translations[value],
        reason = ""
    }
end

-- local function btw12n12(value)
--     local num = tonumber(value)
--     if num == nil then
--         return {
--             receipt = nil,
--             reason = value.." is not a number"
--         }
--     end
--     if num < -12 or num > 12 then
--         return {
--             receipt = nil,
--             reason = "timezone must be within range [-12, 12]."
--         }
--     end
--     return {
--         receipt = num,
--         reason = "success"
--     }
-- end

local interp_funcs = {
    valid_timestr=valid_timestr,
    interpret_bool=interpret_bool
}

local nvdefault_settings = {
    nv_showtime = {
        description = "Toggles whether to display the time in the bottom left",
        possible_values = "true/false",
        value = nil,
        default = true,
        interp = "interpret_bool",
    },
    nv_ampm = {
        description = "Toggles 12hr vs 24hr clock.",
        possible_values = "true/false",
        value = nil,
        default = true,
        interp = "interpret_bool",
    },
    nv_tzoffset = {
        description = "Timezone offset from UTC. Visit https://greenwichmeantime.com/articles/clocks/global/ to get your offset",
        possible_values = "-4, 5:30, 0",
        value = nil,
        default = "-4:00",
        interp = "valid_timestr",
    },
    nv_donottrackme = {
        description = "If enabled, will remove your username from NuVote's popularity tracking.",
        possible_values = "true/false",
        value = nil,
        default = false,
        interp = "interpret_bool",
    },
}

local function nvsetting_func_tostring(tbl)
    return "NuVote Settings object"
end


local function save_player_settings(nvs)
    -- go thru all keys, save only those its value is pop'd
	local fd = io.open(PLAYER_CFGFILE,'w+')

    for k, tbl in pairs(nvs) do
        if type(tbl) == "function" then continue end
        if tbl.value == nil then continue end
        fd:write(k.." "..tostring(tbl.value).."\n")
    end
    fd:flush()
    fd:close()
end

local nvsettings_proto = {}
local function nvs_get(nvs,setting)
    if nvs == nil then
        nvs = cg_deepcopy(nvsettings_proto)
    end
    if nvs[setting] == nil then
        if nvdefault_settings[setting] == nil then
            return nil
        end
        nvs[setting] = nvdefault_settings[setting]
    end

    if nvs[setting].value == nil then
        return nvs[setting].default
    end
    return nvs[setting].value
end

rawset(_G,"nvs_get",nvs_get)

local function nvs_set(nvs,setting,value)
    local mynvs = ""
    if consoleplayer and consoleplayer.nvsettings then
        mynvs = tostring(consoleplayer.nvsettings)
    end
    -- nvdbp("Got nvset "..tostring(nvs).." my nvs: "..mynvs)
    -- I CAN"T F'ING RAISE AN EXCEPTION AAAA
    if nvs[setting] == nil then
        return {
            receipt=nil,
            reason=setting.."is not a valid setting"
        }
    end
    local interp_function = interp_funcs[nvs[setting].interp]
    local interp_result = interp_function(value)
    if interp_result.receipt == nil then
        return interp_result
    end

    nvs[setting].value = interp_result.receipt
    return {
        receipt=interp_result.receipt,
        reason="success"
    }
end

rawset(_G,"nvs_set",nvs_set)

local function nvs_notset(nvs,setting)
    if nvs[setting] == nil then
        if nvdefault_settings[setting] == nil then
            return nil
        end
        nvs[setting] = nvdefault_settings[setting]
    end

    return nvs[setting].value == nil
end

rawset(_G,"nvs_notset",nvs_notset)

for k,setting in pairs(nvdefault_settings) do
    -- setmetatable(setting.interp,{__tostring=nvsetting_func_tostring})
    nvsettings_proto[k] = setting
end

local function __nvs_set_helper(p,setting,value)
    nvdbp("Got helper function p: "..tostring(p))
    if p == nil then return end
    if p.nvsettings == nil then
        p.nvsettings = cg_deepcopy(nvsettings_proto)
    end
    
    if p.nvsettings[setting] == nil then
        nvdbp("tried to set "..setting.."over network, but it didn't exist")
        return
    end

    local interp_function = interp_funcs[p.nvsettings[setting].interp]
    local interp_result = interp_function(value)
    if interp_result.receipt == nil then
        nvdbp("tried to set "..setting.."over network, but it interp didn't work. Message: "..interp_result.reason)
        return interp_result
    end

    p.nvsettings[setting].value = interp_result.receipt
end

COM_AddCommand("__nvs_set_helper",__nvs_set_helper)


local function init_local_settings()
    consoleplayer.nvsettings = cg_deepcopy(nvsettings_proto)
    -- load from NuVotePlayer.cfg:
    local fd = io.open(PLAYER_CFGFILE,'r')
    if fd != nil then
        for line in fd:lines() do
            local data = pipesplit(line," ")
            local key, value = data[1], data[2]
            local result = nvs_set(consoleplayer.nvsettings,key,value)
            if result.receipt == nil then
                CONS_Printf(consoleplayer,"warning: failed to set NuVote setting "..key.." to "..value)
                CONS_Printf(consoleplayer,"Reason: "..result.reason)
            else
                local cmd = string.format("__nvs_set_helper %s %s",key,value)
                COM_BufInsertText(consoleplayer,cmd)
            end
        end
        fd:close()
    end
end

local function client_ready()
    if consoleplayer == nil then return false end
    if consoleplayer.nvsettings == nil then
        init_local_settings()
    end

    return true
end

local function load_settings()
    if not client_ready() return end
end

for k,_ in pairs(nvdefault_settings) do
    local cmdname = k
    COM_AddCommand(cmdname,function(p,value)
        -- if isserver then return end
        -- nvdbp("Got pcvar for"..tostring(p))

        if not client_ready() then
            CONS_Printf(p,"Your client wasn't ready")
        end
        if p.nvsettings == nil then
            if consoleplayer and consoleplayer == p then 
                CONS_Printf(p,"nvsettings is not ready for this player")
                return
            else
                p.nvsettings = cg_deepcopy(nvsettings_proto)
            end
        end
        if value == nil then
            -- print the description for the key
            if p.nvsettings[k] == nil then
                CONS_Printf(p,"No nvsetting exits for this key. (This is an error)")
                return
            end
            local msg = string.format("\"%s\" is \"%s\" default is \"%s\". Possible values [%s]",
            k, tostring(nvs_get(p.nvsettings,k)), tostring(p.nvsettings[k].default), p.nvsettings[k].possible_values)
            CONS_Printf(p,msg)
            CONS_Printf(p,"Description: "..p.nvsettings[k].description)
            return
        end
        -- nvdbp("Hit pcvar cmd",k,value,tostring(p.nvsettings))
        local result = nvs_set(p.nvsettings, k,value)
        -- nvdbp("result of set",result.receipt,result.reason)
        if consoleplayer and consoleplayer == p then
            save_player_settings(p.nvsettings)
        end

        if result.receipt == nil then
            CONS_Printf(p,"Couldn't change setting "..k.." to "..tostring(value))
            CONS_Printf(p,"Reason: "..result.reason)
        end
    end)
end

addHook("ThinkFrame",function()
    if consoleplayer != nil and consoleplayer.nvsettings == nil then
        client_ready()
    end
end)
